#ifndef __MIDIIO__
#define __MIDIIO__

#include <stdio.h>

#ifndef MIDI_BUFSIZE
#define MIDI_BUFSIZE  1024
#endif

const unsigned long MThd = 0x4D546864ul;
const unsigned long MTrk = 0x4D54726Bul;

#define OPTION_NOCONTROLS   1
#define OPTION_NOEVENTS     2

#define NOCHANNEL     (-1)
#define MULTICHANNEL  (-2)

// for use of param what in function text()
#define meta_text       1
#define meta_copyright  2
#define meta_trackname  3
#define meta_instrument 4
#define meta_lyric	5
#define meta_marker	6
#define meta_cuepoint	7

class MidiRead
{
public:
  MidiRead(const char* filename);
  ~MidiRead();

  int run();
  int runhead();
  int runtrack(int trackno);

  long getpos() { return pos_; }

  virtual void head(unsigned songs, unsigned tracks, unsigned clicksperquarter);
  virtual void track(int trackno, unsigned length, int channel);
  virtual void endtrack(int trackno);

  virtual void seqnumber(unsigned int seqno);
  virtual void text(int what, int len, char* whattext, unsigned char* txt);
  virtual void meta(int what, int len, unsigned char* data);
  virtual void end();
  virtual void smpteofs(int hour, int min, int sec, int frame, int fracframe);
  virtual void tact(int nom, int denom, int clicksperbeat, int notes32perbeat);
  virtual void tempo(unsigned long ticks);
  virtual void key(int signature, int isminor); // C=0, -7=7flats +7=7sharps
  virtual void program(int channel, int prg);
  virtual void control(int channel, int what, int val);
  virtual void highbank(int channel, int val);
  virtual void wheel(int channel, int val);
  virtual void breath(int channel, int val);
  virtual void foot(int channel, int val);
  virtual void portamentotime(int channel, int val);
  virtual void data(int channel, int val);
  virtual void volume(int channel, int val);
  virtual void balance(int channel, int val);
  virtual void expression(int channel, int val);
  virtual void lowbank(int channel, int val);
  virtual void hold(int channel, int val);
  virtual void reverb(int channel, int val);
  virtual void chorus(int channel, int val);
  virtual void datainc(int channel, int val);
  virtual void datadec(int channel, int val);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void time(unsigned long ticks);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(int syslen, unsigned char* sysdata);
  virtual void endmidi();
  virtual void error(const char* msg);

  virtual void percent(int perc);

  int getchannel() { return curchannel_; }
  void setchannel(int channel);

  static const char* progname(int n, int channel);
  static const char* notename(unsigned char note);

  int options_;

  int songs_, tracks_, clicks_;

protected:
  const char *midiname_;
  FILE* f_;
  long filesize_;
  unsigned char buf_[MIDI_BUFSIZE];
  int buflen_, bufpos_;
  int curchannel_;
  int percent_;


  long pos_, curpos_;

  unsigned char* need(int nbytes);

  int getbyte();
  unsigned getword();
  unsigned long gettri();
  unsigned long getlong();
  unsigned long getdelta();
  unsigned char* get(int len);
  void seek(long pos);

  int scanchannel(unsigned long maxlen); // channel 0-15 or -1=no channel or -2=multichannels
  int scansysevent(unsigned long maxlen);
};

class MidiWrite
{
public:
  MidiWrite(const char* filename);
  ~MidiWrite();

  void head(unsigned clicksperquarter = 192);
  void track();
  void endtrack();

  void event(int what, int len, unsigned char* data);

  void text(int what, int len, unsigned char* txt);
  void meta(int what, int len, unsigned char* data); // 0xff ....
  void end();
  void tact(int nom, int denom, int v1, int v2);
  void tempo(unsigned long ticks);
  void program(int channel, int prg);
  void control(int channel, int what, int val);
  void highbank(int channel, int val);
  void wheel(int channel, int val);
  void breath(int channel, int val);
  void foot(int channel, int val);
  void portamentotime(int channel, int val);
  void data(int channel, int val);
  void volume(int channel, int val);
  void balance(int channel, int val);
  void expression(int channel, int val);
  void lowbank(int channel, int val);
  void hold(int channel, int val);
  void reverb(int channel, int val);
  void chorus(int channel, int val);
  void datainc(int channel, int val);
  void datadec(int channel, int val);
  void noteon(int channel, int note, int vel);
  void noteoff(int channel, int note, int vel);
  void time(unsigned long ticks);
  void pitchbend(int channel, int val);
  void polyaftertouch(int channel, int note, int val);
  void aftertouch(int channel, int val);
  void songpos(unsigned pos);
  void songselect(unsigned char song);
  void tunerequest();
  void timingclock();
  void start();
  void cont();
  void stop();
  void activesense();
  void sysex(int syslen, unsigned char* sysdata);

  void putbyte(unsigned char val);
  void putcode(unsigned char code);
  void putword(unsigned val);
  void puttri(unsigned long val);
  void putlong(unsigned long val);
  void putdelta(unsigned long val);
  void puttime();
  void put(int len, unsigned char* c);
  void seek(long pos);

protected:
  const char *midiname_;
  FILE* f_;
  long trackpos_, curpos_, filesize_;
  int trackchannel_, trackcount_, lastcode_, endtrack_;

  unsigned char buf_[MIDI_BUFSIZE];
  int bufpos_, buflen_;

  unsigned long curdelta_;

  void flush();
};

#endif
